const zlib = require('zlib');
const { promisify } = require('util');

const gzip = promisify(zlib.gzip);
const gunzip = promisify(zlib.gunzip);
const brotliCompress = promisify(zlib.brotliCompress);
const brotliDecompress = promisify(zlib.brotliDecompress);

class DataCompressor {
    constructor(method) {
        this.method = method || process.env.COMPRESSION_METHOD;
    }

    async compress(data) {
        const bufferData = Buffer.from(data, 'utf8');
        switch (this.method) {
            case 'GZIP':
                return await gzip(bufferData);
            case 'BROTLI':
                return await brotliCompress(bufferData);
            default:
                return bufferData;  // No compression
        }
    }

    async decompress(compressedData) {
        try {
            let result;
            switch (this.method) {
                case 'GZIP':
                    result = await gunzip(compressedData);
                    break;
                case 'BROTLI':
                    result = await brotliDecompress(compressedData);
                    break;
                default:
                    result = compressedData;  // No decompression
            }
            return result.toString('utf8');
        } catch (error) {
            if (error.code === 'Z_DATA_ERROR') {
                process.log.error('Data decompression error. Data might not be compressed or might be corrupted.');
                return compressedData.toString('utf8');  // Return the original data as string or provide other fallback mechanisms.
            }
            throw error;  // Re-throw the error if it's not the one we're handling.
        }
    }
}

module.exports = DataCompressor;