const protobuf = require('protobufjs');
const ROOTprotobuf = protobuf.Root;

const { GQLClient, GQLProfileQuery } = require('@lib/qgl_client.js');

const blockInfoKeys = {
    longestGameChampId: true,
    longestGameDuration: true,
    longestGameTimestamp: true,
    longestKillingSpree: true,
    longestTimeSpentLiving: true,
    longestTimeSpentLivingChampId: true,
    longestKillingSpree: true,
    mostDamage: true,
    mostDamageChampId: true,
    mostDamageTimestamp: true
};

const blockOverallKeys = {
    champId: true,
    keystoneId: true,
    itemId: true
};


let globalLoadedRoot;

/**
 * Prepare the protobuf root for the proto files
 */
const prepareRoot = async () => {
    const root = new ROOTprotobuf();
    root.resolvePath = (origin, target) => {
        return target;
    };

    const response = await fetch(cupcake("profile/proto.belt", true, false));
    const protoContent = await response.text();

    const loadedRoot = protobuf.parse(protoContent, root).root;
    process.log.system("Proto belt ROOT loaded");
    globalLoadedRoot = loadedRoot;
}

/**
 * Get a cupcake URL - reduce code size
 * @param {string} file 
 * @param {boolean} basePath - whether the file is in /2022 or not
 * @param {boolean} isImage - append a .webp to the path
 */
const cupcake = (file, basePath = true, isImage = true) => {
    let path = process.env.CUPCAKE_URL;
    if (basePath) path += `${process.env.CUPCAKE_YEAR}/`;
    path += file;
    if (isImage) path += ".webp";
    process.log.debug(path);
    return path;
}

const queueDataToObject = (queuedata) => {
    let result = {};
    for (let index = 0; index < queuedata.length; index++) {
        const queue = queuedata[index];
        result[queue.queueGroupId ?? index] = queue;
    }

    return result;
}

const arrayToObject = (list) => {
    let result = {};
    for (let index = 0; index < list.length; index++) {
        const item = list[index];
        const { keystoneId, champId = "0", itemId, position = "0" } = item;

        let key;
        if (keystoneId) {
            key = `${keystoneId}:${champId}:${position}`;
        } else if (itemId) {
            key = `${itemId}:${champId}:${position}`;
        } else {
            key = champId;
        }

        result[key] = item;
    }

    return result;
}

const deepClone = (obj) => {
    if (obj === null || typeof obj !== "object") {
        return obj;
    }
    if (Array.isArray(obj)) {
        return obj.map(deepClone);
    }
    const clonedObj = {};
    for (let key of Object.keys(obj)) {
        clonedObj[key] = deepClone(obj[key]);
    }
    return clonedObj;
}

/**
 * @typedef {Object} RawSummoner
 * @property {string} puuid - The unique player identifier.
 * @property {string} server - The server on which the player is located.
 */

/**
 * @typedef {Object} BurstFireSummonerProfile
 * @property {string} name - The name of the summoner.
 * @property {number} profileIconId - The ID of the profile icon.
 * @property {string} s3Url - The URL pointing to the S3 bucket containing the profile's assets.
 * @property {boolean} protected - Flag indicating whether the profile is protected.
 * @property {string} __typename - The GraphQL typename.
 */

/**
 * @typedef {RawSummoner & BurstFireSummonerProfile} CombinedSummonerProfile
 */


/**
 * 
 * @param {BurstFireSummonerProfile[]} summoner 
 * @returns {}
 */
const asyncLoadProfiles = async (summoner) => {
    try {

        const SummonerProfile = globalLoadedRoot.lookupType("shockwave.SummonerProfile");

        for (let profile of summoner) {
            // Should use s3URL from GQL
            const response = await fetch(`https://protobelt.player-val.eu/file/yil-2023/${profile.puuid}`);

            if (response.status !== 200) {
                profile.status_code = response.status;
                throw new Error(`Status code ${response.status}`);
            } else {
                const buffer = await response.arrayBuffer();
                const decoded = SummonerProfile.decode(new Uint8Array(buffer));
                profile.data = decoded;
            }
        }

        return summoner;

    } catch (err) {
        console.error(err);
        // Potentially handle more robust error logging or action here
    }
};

/**
 * Fetch a array of summoner profiles from the GraphQL API.
 * @param {RawSummoner[]} summoner 
 * @returns {BurstFireSummonerProfile[]}
 */
const fetchSummoner = async (summoner) => {
    const query = GQLProfileQuery(summoner);

    /** @type {BurstFireSummonerProfile[]} */
    const request = await GQLClient.query(query).toPromise();

    if (request.error || typeof request.data === "undefined") {
        console.error(request);
        throw new Error(`General <SUMMONER> GraphQL request failed`);
    }

    // Check if the data of the keys in the data object are null
    for (const key in request.data) {
        if (Object.hasOwnProperty.call(request.data, key)) {
            const element = request.data[key];
            const index = parseInt(key.replace("profile", ""));

            if (element === null) {
                throw new Error(`GraphQL request failed on ${summoner[index].puuid}:${summoner[index].shard}`);
            }
        }
    }

    return request.data;
}

const mergeProtos = (oldData) => {

    if (oldData.length === 1) {
        return deepClone(oldData)[0];
    }

    let pentakills = [];
    // got rid of the reference
    const data = oldData.map((group) => {

        if (typeof group.pentaKills !== "undefined") {
            pentakills = pentakills.concat(group.pentaKills);
        }

        // this re-assignment is required. Dunno why, but otherwise it breaks 
        return queueDataToObject(deepClone(group.data));
    });

    const response = {};

    // loop through the players
    for (let i = 0; i < data.length; i++) {
        const player = data[i];


        // loop through the queue groups 
        const queueGroups = Object.values(player);
        for (let j = 0; j < queueGroups.length; j++) {
            const queueGroup = queueGroups[j];
            const queueGroupId = queueGroup.queueGroupId ?? 0;

            if (!response[queueGroupId]) {
                response[queueGroupId] = { queueGroupId };
            }

            // loop through info, champs, keystones and items
            const keys = Object.keys(queueGroup);
            for (let k = 0; k < keys.length; k++) {
                const categoryName = keys[k];

                // skip queueGroupId
                if (categoryName === "queueGroupId") {
                    continue;
                }
                const category = queueGroup[categoryName];

                if (!response[queueGroupId][categoryName]) {
                    response[queueGroupId][categoryName] = {};
                }

                if (categoryName === "info") {
                    for (const key in category) {
                        if (Object.hasOwnProperty.call(category, key)) {
                            const element = category[key];

                            if (!response[queueGroupId][categoryName][key]) {
                                response[queueGroupId][categoryName][key] = element;
                                continue;
                            }

                            // skip forbidden keys
                            if (blockInfoKeys.hasOwnProperty(key)) {

                                if (key === "longestGameDuration" && element > response[queueGroupId][categoryName][key]) {
                                    response[queueGroupId][categoryName][key] = element;
                                    response[queueGroupId][categoryName]["longestGameChampId"] = category["longestGameChampId"] ?? 1;
                                    response[queueGroupId][categoryName]["longestGameTimestamp"] = category["longestGameTimestamp"] ?? 1;
                                    continue;
                                }

                                if (key === "mostDamage" && element > response[queueGroupId][categoryName][key]) {
                                    response[queueGroupId][categoryName][key] = element;
                                    response[queueGroupId][categoryName]["mostDamageChampId"] = category["mostDamageChampId"] ?? 1;
                                    response[queueGroupId][categoryName]["mostDamageTimestamp"] = category["mostDamageTimestamp"] ?? 1;
                                    continue;
                                }

                                if (key === "longestKillingSpree" && element > response[queueGroupId][categoryName][key]) {
                                    response[queueGroupId][categoryName][key] = element;
                                    continue;
                                }

                                if (key === "longestTimeSpentLiving" && element > response[queueGroupId][categoryName][key]) {
                                    response[queueGroupId][categoryName][key] = element;
                                    response[queueGroupId][categoryName]["longestTimeSpentLivingChampId"] = category["longestTimeSpentLivingChampId"] ?? 1;
                                    continue;
                                }

                                continue;
                            }

                            if (typeof element === "number") {
                                response[queueGroupId][categoryName][key] += element;
                                continue;
                            }

                            if (typeof element === "object") {
                                if (Array.isArray(element)) {
                                    response[queueGroupId][categoryName][key].push(...element);
                                    continue;
                                }

                                for (const objectKey in element) {
                                    if (Object.hasOwnProperty.call(element, objectKey)) {
                                        const elementObject = element[objectKey];

                                        if (!response[queueGroupId][categoryName][key][objectKey]) {
                                            response[queueGroupId][categoryName][key][objectKey] = elementObject;
                                            continue;
                                        }

                                        response[queueGroupId][categoryName][key][objectKey] += elementObject;
                                    }
                                }
                            }
                        }
                    }
                    continue;
                }

                const parsedContent = arrayToObject(category);

                for (const key in parsedContent) {
                    if (Object.hasOwnProperty.call(parsedContent, key)) {
                        // element is either item, champ or keystone
                        const element = parsedContent[key];

                        if (!response[queueGroupId][categoryName][key]) {
                            response[queueGroupId][categoryName][key] = {};
                        }

                        for (const item in element) {
                            if (Object.hasOwnProperty.call(element, item)) {
                                const value = element[item];

                                if (!response[queueGroupId][categoryName][key][item]) {
                                    if (item === "teamDamagePct" || item === "teamDamageTakenPct") {
                                        response[queueGroupId][categoryName][key][item] = value * (element.totalGames ?? 0);
                                    } else {
                                        response[queueGroupId][categoryName][key][item] = value;
                                    }
                                    continue;
                                }

                                if (blockOverallKeys.hasOwnProperty(item)) {
                                    continue;
                                }

                                if (typeof value === "number") {
                                    if (item === "teamDamagePct" || item === "teamDamageTakenPct") {
                                        response[queueGroupId][categoryName][key][item] += value * (element.totalGames ?? 1);

                                        if (i === data.length - 1) {
                                            response[queueGroupId][categoryName][key][item] = response[queueGroupId][categoryName][key][item] / (response[queueGroupId][categoryName][key].totalGames ?? 1);
                                        }
                                        continue;
                                    }
                                    response[queueGroupId][categoryName][key][item] += value;
                                    continue;
                                }

                                // TODO: we have to merge the lanes somehow :pepeCross:
                                if (typeof value === "object" && !Array.isArray(value)) {
                                    for (const objectKey in value) {
                                        if (Object.hasOwnProperty.call(value, objectKey)) {
                                            const elementObject = value[objectKey];

                                            if (!response[queueGroupId][categoryName][key][item][objectKey]) {
                                                response[queueGroupId][categoryName][key][item][objectKey] = elementObject;
                                                continue;
                                            }

                                            response[queueGroupId][categoryName][key][item][objectKey] += elementObject;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    const returnContentFull = deepClone(oldData[0]);
    returnContentFull.data = Object.values(response);

    for (let i = 0; i < returnContentFull.data.length; i++) {

        returnContentFull.data[i]["champs"] = Object.values(returnContentFull.data[i]["champs"] ?? {});
        returnContentFull.data[i]["items"] = Object.values(returnContentFull.data[i]["items"] ?? {});
        returnContentFull.data[i]["keystones"] = Object.values(returnContentFull.data[i]["keystones"] ?? {});

    }

    returnContentFull.pentaKills = pentakills;

    return returnContentFull;
};

const extractProtoData = (protoData) => {

    const clonedData = deepClone(protoData);

    const response = {
        totalGames: 0,
        totalWins: 0,
        totalLosses: 0,
        totalSeconds: 0,
        totalGoldEarned: 0,
        totalKills: 0,
        totalDeaths: 0,
        totalAssists: 0,
        totalCS: 0,
        totalDamage: 0,
        totalSecondsSpentDead: 0,
        totalButtons: 0,
        totalPentaKills: 0,
    };

    const pentas = protoData.pentaKills ?? [];
    const champions = {};

    const roles = {
        top: { "games": 0, "wins": 0 },
        jgl: { "games": 0, "wins": 0 },
        mid: { "games": 0, "wins": 0 },
        bot: { "games": 0, "wins": 0 },
        sup: { "games": 0, "wins": 0 },
        aram: { "games": 0, "wins": 0 }
    };

    const data = clonedData.data;

    for (let i = 0; i < data.length; i++) {
        const queueGroup = data[i];
        const info = queueGroup.info;

        response.totalGames += info.totalGames ?? 0;
        response.totalWins += info.totalWins ?? 0;

        response.totalSeconds += info.totalSeconds ?? 0;
        response.totalSecondsSpentDead += info.timeSpentDead ?? 0;

        response.totalGoldEarned += info.goldEarned ?? 0;

        response.totalKills += info.totalKills ?? 0;
        response.totalDeaths += info.totalDeaths ?? 0;
        response.totalAssists += info.totalAssists ?? 0;

        response.totalDamage += (info.physicalDamage ?? 0) + (info.magicalDamage ?? 0) + (info.trueDamage ?? 0);

        // ARAM STUFF
        if (queueGroup.queueGroupId === 2) {
            roles.aram.games += info.totalGames ?? 0;
            roles.aram.wins += info.totalWins ?? 0;
        }

        response.totalButtons += (info.summonerSpellsPressed ?? 0) + (info.buttonsPressed ?? 0);

        const champs = arrayToObject(queueGroup.champs);
        for (const champId in champs) {
            if (Object.hasOwnProperty.call(champs, champId)) {
                const champ = champs[champId];

                // handle the special cases
                if (champ.teamDamagePct) {
                    champ.teamDamagePct *= champ.totalGames ?? 1;
                }

                if (champ.teamDamageTakenPct) {
                    champ.teamDamageTakenPct *= champ.totalGames ?? 1;
                }

                champ.pentaKills = 0;


                if (!champions[champId]) {
                    champions[champId] = champ;
                    continue;
                }
                for (const key in champ) {
                    if (Object.hasOwnProperty.call(champ, key)) {
                        const value = champ[key];

                        if (!champions[champId][key]) {
                            champions[champId][key] = value;
                            continue;
                        }
                        if (key === "champId") continue;
                        if (typeof value === "number") {
                            champions[champId][key] += value;
                            continue;
                        }

                        if (typeof value === "object" && !Array.isArray(value)) {
                            for (const k in value) {
                                if (Object.hasOwnProperty.call(value, k)) {
                                    const insert = value[k];
                                    if (!champions[champId][key][k]) {
                                        champions[champId][key][k] = insert;
                                        continue;
                                    }
                                    if (typeof insert === "number") {
                                        champions[champId][key][k] += insert;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    for (let i = 0; i < pentas.length; i++) {
        const penta = pentas[i];

        if (champions[penta.champId]) {
            champions[penta.champId].pentaKills += penta.count;
            response.totalPentaKills += penta.count;
        }
    }

    response.champions = Object.values(champions).map((champ) => {
        if (champ.teamDamagePct) {
            champ.teamDamagePct = champ.teamDamagePct / (champ.totalGames ?? 1);
        }

        if (champ.teamDamageTakenPct) {
            champ.teamDamageTakenPct = champ.teamDamageTakenPct / (champ.totalGames ?? 1);
        }

        if (champ.cs) {
            response.totalCS += champ.cs ?? 0;
        }

        response.totalButtons += (champ.qCasts ?? 0) + (champ.wCasts ?? 0) + (champ.eCasts ?? 0) + (champ.rCasts ?? 0);

        if (champ.positions) {
            for (const position in champ.positions) {
                if (Object.hasOwnProperty.call(champ.positions, position)) {
                    roles[position]["games"] += champ.positions[position];
                }
            }

            for (const position in champ.totalWinsLane) {
                if (Object.hasOwnProperty.call(champ.totalWinsLane, position)) {
                    roles[position]["wins"] += champ.totalWinsLane[position];
                }
            }
        }

        return champ;
    });
    response.champions.sort((a, b) => {

        if (a.totalGames === b.totalGames) {
            return b.totalSeconds - a.totalSeconds;
        }

        return b.totalGames - a.totalGames;
    });

    response.roles = roles;

    response.totalLosses = response.totalGames - response.totalWins;

    return response;

};


module.exports = {
    deepClone,
    cupcake,
    prepareRoot,
    mergeProtos,
    extractProtoData,
    fetchSummoner,
    asyncLoadProfiles
}
