const WebSocket = require('ws');
const { CommunicationError } = require('@lib/errors');
const AppPerformanceMonitor = require('@lib/app_stats');

class WebSocketManager {
    constructor(ws_protocol, target, port) {
        this.AppPerformanceMonitor = new AppPerformanceMonitor();
        this.ws_protocol = ws_protocol;
        this.target = target;
        this.port = port;

        // Reconnect Variable
        /** @const {Number} [reconnectInterval] */
        this.reconnectInterval = 1000;

        // WS Connection stuff
        this.ws = null;
        this.firstConnect = true;
        this.tokenValid = false;

        // Reconnect Stuff
        this.failedReconnectAttempts = 0;
        this.wasareconnect = false;
        this.reconnectAttemps = process.env.RECONNECTATTEMPS || 3;

        // puppeteer
        this.puppeteer = null;

        // Avaible States
        /**
         * @typedef {Array} com_states
         * @property {String} starting
         * @property {String} benchmarking
         * @property {String} ready
         * @property {String} shutting_down
         */
        this.com_states = [
            "starting",
            "benchmarking",
            "ready",
            "shutting_down"
        ]

        process.eventbus.on('reconnect', async () => {
            if (this.failedReconnectAttempts < this.reconnectAttemps) {
                setTimeout(async () => {
                    this.wasareconnect = true;
                    process.log.info(`Trying to reconnect... (${this.failedReconnectAttempts + 1}/${this.reconnectAttemps})`)
                    this.failedReconnectAttempts++
                    await this.connect();
                }, this.reconnectInterval * this.failedReconnectAttempts)
            } else {
                process.log.warning('Connection could not be rebuild. A send will still try again!')
            }
        });

        setInterval(this.#innerLoop, 1000);
    }

    #innerLoop = () => {
        if (this.ws && this.ws.readyState === WebSocket.OPEN) {
            // Only send stats when we setup puppeteer
            if (this.puppeteer) {
                const stats = {
                    "app": this.AppPerformanceMonitor.getStats(),
                    "puppeteer": {
                        "queue": this.puppeteer.getCluster.jobQueue.list.length ?? 0,
                        "chrome": this.#calcAverage(this.puppeteer.chrome_memory_usage) ?? 0,
                        "cdn_cache_size": this.#calcAverage(this.puppeteer.getimgCache_size) ?? 0,
                        "cdn_served_mebibytes": this.#calcAverage(this.puppeteer.getserved_mebibytes),
                        "time_latancy": Math.max(this.#calcAverage(this.puppeteer.getscreenshot_rendertimes) || 0, 200),
                        "render_success": this.puppeteer.getscreenshot_renderd,
                        "render_failed": this.puppeteer.getscreenshot_failed,
                        "estimated_performance_perSecond": this.#calcAverage(this.puppeteer.getrunsPerS_array) ?? 0,
                    }
                }
                this.ws.send(JSON.stringify({ "com": "send:stats", "payload": stats }));
            }
        }
    }

    /**
     * Calculate the average of an array
     * @param {Array} array 
     * @returns 
     */
    #calcAverage = (array) => {
        const sum = array.reduce((a, b) => a + b, 0);
        return sum / array.length;
    }

    connect() {
        return new Promise((resolve, reject) => {
            this.ws = new WebSocket(`${this.ws_protocol}://${this.target}:${this.port}/api/v1/websocket/con`);
            this.ws.on('open', () => {
                process.log.info('Connected to server');
                this.failedReconnectAttempts = 0;
                // Authendicate websocket
                this.ws.send(JSON.stringify({
                    "com": "auth",
                    "payload": process.env.BALANCER_TOKEN
                }));
            });

            this.ws.on('message', (msg) => {
                process.log.debug(`Received: ${msg}`);
                const { com, payload } = JSON.parse(msg);
                if (com === 'auth') {
                    if(payload){ 
                        this.tokenValid = true
                    } else {
                        this.tokenValid = false
                        this.ws.close()
                    }
                    if (this.firstConnect) {
                        resolve(true)
                        this.firstConnect = false;
                    } else {
                        // We need to send config if its a reconnect because we already benchmarked
                        this.ws.send(JSON.stringify({
                            "com": "get:config",
                            "payload": {
                                "me": process.me,
                                "cuncurrency": this.puppeteer.getMAX_CONCURRENCY,
                            }
                        }))
                    }
                }
                if (com === 'config') {
                    this.puppeteer.setConfig(payload);
                    this.puppeteer.setSend(this.send);
                    // Make sure the balancer already got its first metrics ping
                    setTimeout(() => {
                        this.setState('ready');
                    }, 1000);
                }

                if(com === 'request:wipeCache'){
                    this.puppeteer.wipecache()
                }

                if(com === 'request:drainAndKill'){
                    process.log.warning('Drain and Kill requested. Shutting down...')
                    process.emit('SIGTERM')
                }

                if (com === 'request:pic') {
                    const { slug, puuid_array, lang, design } = payload
                    this.puppeteer.addQueue(slug, puuid_array, lang, design)
                }

                if (com === 'message:info') {
                    process.log.info(payload)
                }

                if (com === 'message:warn') {
                    process.log.warning(payload)
                }

                if (com === 'message:alert') {
                    process.log.error(payload)
                }
            });

            this.ws.on('close', async () => {
                process.log.info('Connection closed');
                if (!this.tokenValid) process.log.error('Invalid token. Please check your .env file and restart the application.');
                if (!this.tokenValid) process.exit(1800)
                this.ws = null;
                if (this.failedReconnectAttempts < this.reconnectAttemps) {
                    setTimeout(async () => {
                        this.wasareconnect = true;
                        process.log.info(`Trying to reconnect... (${this.failedReconnectAttempts + 1}/3)`)
                        this.failedReconnectAttempts++
                        await this.connect();
                    }, this.reconnectInterval * this.failedReconnectAttempts)
                }
            });
        });
    }

    send = async (com, payload, retry_max = 0) => {
        let retry = 0;
        if (this.ws && this.ws.readyState === WebSocket.OPEN) {
            this.ws.send(JSON.stringify({ "com": com, "payload": payload }));
        } else {
            await this.connect();
            process.log.warning(`Could not send data to server. (${retry + 1}) Retrying...`)
            if (retry < retry_max) {
                setTimeout(() => {
                    retry++;
                    this.send(com, payload, retry_max);
                }, 100);
            } else {
                throw new CommunicationError('Could not send data to server')
            }
        }
    }

    setPuppeteer(puppeteer) {
        this.puppeteer = puppeteer;
        // Request config (To initiate the node)
        this.ws.send(JSON.stringify({
            "com": "get:config",
            "payload": {
                "me": process.me,
                "cuncurrency": this.puppeteer.getMAX_CONCURRENCY,
            }
        }))
    }

    /**
     * Update the current state of the node
     * @param {com_states} state 
     */
    setState(state) {
        if (this.com_states.includes(state)) {
            this.send("send:state", {
                "state": state,
                "me": process.me
            })
        } else {
            process.log.error(`State ${state} is not avaible!`)
        }
    }
}

module.exports = WebSocketManager;