const { Cluster } = require('puppeteer-cluster');
const path = require('path');
const fs = require('fs');
const { GenericPuppeteerClusterError, PuppeteerClusterError } = require('@lib/errors');
const { asyncLoadProfiles, fetchSummoner, extractProtoData, mergeProtos, cupcake } = require('@lib/proto.js');
const { getserved_bytes, getChampion, getSkinById, beautifySecondsToFixedString, template, CSStemplate, queueGroupMerger, clearimageCache, shortenNumberByLocale, getWebPImageAsBase64, getProfileWebPImageAsBase64, getCacheUsageInBytes } = require('@lib/card_utils.js');
const { exec } = require('child_process');
const zlib = require('zlib');
const os = require('os');
const performance = require('perf_hooks').performance;

const minimal_args = [
    '--autoplay-policy=user-gesture-required',
    '--disable-background-networking',
    '--disable-background-timer-throttling',
    '--disable-backgrounding-occluded-windows',
    '--disable-breakpad',
    '--disable-client-side-phishing-detection',
    '--disable-component-update',
    '--disable-default-apps',
    '--disable-dev-shm-usage',
    '--disable-domain-reliability',
    '--disable-extensions',
    '--disable-features=AudioServiceOutOfProcess',
    '--disable-hang-monitor',
    '--disable-ipc-flooding-protection',
    '--disable-notifications',
    '--disable-offer-store-unmasked-wallet-cards',
    '--disable-popup-blocking',
    '--disable-print-preview',
    '--disable-prompt-on-repost',
    '--disable-renderer-backgrounding',
    '--disable-setuid-sandbox',
    '--disable-speech-api',
    '--disable-sync',
    '--hide-scrollbars',
    '--ignore-gpu-blacklist',
    '--metrics-recording-only',
    '--mute-audio',
    '--no-default-browser-check',
    '--no-first-run',
    '--no-pings',
    '--no-sandbox',
    '--no-zygote',
    '--password-store=basic',
    '--use-gl=swiftshader',
    '--use-mock-keychain',
];

class PuppeteerClusterManager {
    constructor(MAX_CONCURRENCY) {
        this.MAX_CONCURRENCY = MAX_CONCURRENCY || 1

        this.cluster = null;
        this.config = {};

        // All Stores to report the 60s average
        this.screenshot_rendertimes = [];
        this.chrome_memory_usage = [];
        this.imgCache_size = [];
        this.served_mebibytes = [];
        this.runsPerS_array = [];

        // Other Stats
        this.screenshot_renderd = 0;
        this.screenshot_failed = 0;

        // Data to guess the 60s average of the max amount of screenshots possible
        this.screenshot_counter = 0;
        this.active_time_passed = 0;
        this.lase_activeTimeStamp = Date.now();
        this.ScreenshotRendering = false;

        // Card Template
        this.card_templates = {
            "dark": {
                "index": CSStemplate(fs.readFileSync(path.join(__dirname, '..', 'cards', 'dark', 'build', 'index.html'), 'utf8'), {
                    "css": `<style>${fs.readFileSync(path.join(__dirname, '..', 'cards', 'dark', 'build', 'card.css'), 'utf8')}</style>`,
                    "yil_logo": fs.readFileSync(path.join(__dirname, '..', 'cards', 'yil-logo.svg'), 'utf8'),
                }),
                "champ_full": fs.readFileSync(path.join(__dirname, '..', 'cards', 'dark', 'build', 'championFull.html'), 'utf8'),
                "champ_empty": fs.readFileSync(path.join(__dirname, '..', 'cards', 'dark', 'build', 'championEmpty.html'), 'utf8'),
            },
            "alternative": {
                "index": CSStemplate(fs.readFileSync(path.join(__dirname, '..', 'cards', 'alternative', 'build', 'index.html'), 'utf8'), {
                    "css": `<style>${fs.readFileSync(path.join(__dirname, '..', 'cards', 'alternative', 'build', 'card.css'), 'utf8')}</style>`,
                    "yil_logo": fs.readFileSync(path.join(__dirname, '..', 'cards', 'yil-logo.svg'), 'utf8'),
                }),
                "arena": fs.readFileSync(path.join(__dirname, '..', 'cards', 'alternative', 'build', 'arena.html'), 'utf8'),
                "nexus": fs.readFileSync(path.join(__dirname, '..', 'cards', 'alternative', 'build', 'nexus.html'), 'utf8'),
                "urf": fs.readFileSync(path.join(__dirname, '..', 'cards', 'alternative', 'build', 'urf.html'), 'utf8'),
                "not_played": fs.readFileSync(path.join(__dirname, '..', 'cards', 'alternative', 'build', 'notplayed.html'), 'utf8'),
                "champions": fs.readFileSync(path.join(__dirname, '..', 'cards', 'alternative', 'build', 'champions.html'), 'utf8'),
            }
        }

        this.some_Interval = setInterval(async () => {
            process.log.debug(`Cards counter: ${this.screenshot_counter} | Active time passed: ${this.active_time_passed} | Card rendering: ${this.ScreenshotRendering}`);
            const runsPerS = (((this.screenshot_counter / this.active_time_passed) * 1000).toFixed(2)); // When not rendering, it devides by 0 for fun :D
            // Check if runsPerS is a number
            if (!isNaN(runsPerS)) {
                this.runsPerS_array.push(parseFloat(runsPerS));
            }
            // Reset counters
            this.active_time_passed = 0;
            this.screenshot_counter = 0;
            let mem = await this.findChromiumProcesses();
            this.chrome_memory_usage.push(mem.reduce((sum, item) => sum + parseFloat(item.memUsage), 0));
            this.imgCache_size.push(getCacheUsageInBytes());
            this.served_mebibytes.push(getserved_bytes());
            // Make sure we only have 60 entries
            this.#shiftArrayAndCut(this.chrome_memory_usage);
            this.#shiftArrayAndCut(this.runsPerS_array);
            this.#shiftArrayAndCut(this.imgCache_size);
            this.#shiftArrayAndCut(this.served_mebibytes);
        }, 1000);
    }

    /**
     * Calculate the average of an array
     * @param {Array} array 
     * @returns 
     */
    #calcAverage = (array) => {
        const sum = array.reduce((a, b) => a + b, 0);
        return sum / array.length;
    }

    /**
     * Set the config for the cluster (got from the balancer)
     * @param {Object} config 
     */
    setConfig(config) {
        // We need to verify if there was already a config, if so check if balancer name is the same
        // If the name is the same, we can just update the config
        // If the name is different, we need to flush all data (Means the balancer died and lost all connections)
        if (Object.keys(this.config).length !== 0) {
            if (this.config.BALANCER_LIFE_NAME !== config.BALANCER_LIFE_NAME) {
                process.log.warning('Balancer name changed. Flushing all data.')
                //this.cluster.queue.clear();
            }
        }
        this.config = config;
        process.log.info('Puppeteer config updated');
    }

    /**
     * Make the conncetion available to the cluster
     * @param {WebsocketSendFuntion} ws 
     */
    setSend(send) {
        process.log.debug('Puppeteer cluster got new websocket');
        this.send = send;
    }

    #transmitImageBuffer = async (slug, image_buffer) => {
        const start_time = performance.now();
        zlib.gzip(image_buffer, (err, gzipBuffer) => {
            if (err) {
                console.error('Compression error:', err);
                return;
            }
            const end_time = performance.now();
            const time_taken = end_time - start_time;
            process.log.debug(`Compression took ${time_taken.toFixed(3)}ms`);

            const compressionPercentage = (1 - gzipBuffer.length / image_buffer.length) * 100;
            process.log.debug(`Reduced image by ${compressionPercentage.toFixed(2)}%`);
            this.send("send:pic", {
                "slug": slug,
                "image_buffer": gzipBuffer
            });
        });
    }

    /**
     * Initialize the cluster
     */
    async init(benchmarking = false) {
        return new Promise(async (resolve, reject) => {
            this.isBenchmarking = benchmarking;
            if (this.cluster) {
                throw new GenericPuppeteerClusterError('Puppeteer cluster already initialized');
            }

            try {
                const puppeteerArgs = {
                    headless: "new",
                    userDataDir: path.join(__dirname, '..', 'puppeteer-data'),
                    args: minimal_args
                }

                if (process.env.CONTAINER_ARM64 === "production") puppeteerArgs.executablePath = "/usr/bin/chromium";

                this.cluster = await Cluster.launch({
                    concurrency: Cluster.CONCURRENCY_CONTEXT,
                    maxConcurrency: this.MAX_CONCURRENCY,
                    puppeteerOptions: puppeteerArgs
                });

                this.cluster.on('taskerror', (err, data, willRetry) => {
                    if (willRetry) {
                        console.warn(`Encountered an error while crawling ${data}. ${err.message}\nThis job will be retried`);
                    } else {
                        console.error(`Failed to crawl ${data}: ${err.message}`);
                    }
                });

                await this.cluster.task(async ({ page, data }) => {
                    try {
                        const time_start = Date.now();
                        //console.log("TASK DATA", data)

                        await page.setViewport({
                            width: parseInt(this.config.PUPPETEER_SCREENSHOT_WIDTH, 10) || 1080,
                            height: parseInt(this.config.PUPPETEER_SCREENSHOT_HEIGHT, 10) || 1920,
                        });

                        //HTML Data is in data.html_data
                        const start_render_time = Date.now();
                        //await page.goto("http://localhost", { waitUntil: 'domcontentloaded' });
                        await page.setContent(data.html_data, { waitUntil: 'domcontentloaded' }); // networkidle0
                        //await page.goto('https://www.google.com')
                        const imageBuffer = await page.screenshot({
                            type: 'jpeg',
                            quality: parseInt(this.config.PUPPETEER_SCREENSHOT_QUALITY, 10) || 80,
                            clip: {
                                x: 0,
                                y: 0,
                                width: parseInt(this.config.PUPPETEER_SCREENSHOT_WIDTH, 10) || 1080,
                                height: parseInt(this.config.PUPPETEER_SCREENSHOT_HEIGHT, 10) || 1920,
                            },
                            omitBackground: true,
                        });
                        // We end tracking here because we go back to the async part
                        const end_render_time = Date.now();
                        this.active_time_passed += end_render_time - start_render_time;

                        if (!this.isBenchmarking) {
                            this.#transmitImageBuffer(data.slug, imageBuffer)
                        }

                        //fs.writeFileSync(path.join(__dirname, '..', 'www-public', `screenshot-${Number(Math.random(5)) + Number(data.id)}.jpg`), imageBuffer);
                        if (!this.isBenchmarking) {
                            process.log.info(`Card took ${Date.now() - time_start}ms to render`);
                            this.screenshot_renderd++;
                            this.screenshot_counter++
                            this.screenshot_rendertimes.push(Date.now() - time_start);
                            this.#shiftArrayAndCut(this.screenshot_rendertimes);
                            this.ScreenshotRendering = false;
                        }
                    } catch (error) {
                        this.screenshot_failed++;
                        this.screenshot_counter++;

                        this.send("send:pic_err", {
                            "slug": data.slug,
                            "error": "General image Generation Error"
                        });
                    }
                });

                resolve();

            } catch (error) {
                reject(error);
                throw new PuppeteerClusterError(error); // Maybe this leads to a crash but the error should exit the process anyway
            }
        });
    }

    /**
     * Generade a card based on a PUUID array
     * @param {String} slug
     * @param {import('@lib/proto.js').RawSummoner[]} puuid_array
     * @param {String} lang
     * @param {String} design
     */
    addQueue = async (slug, puuid_array, lang, design) => {
        try {
            // If we benchmark, we can´t render real images because we are missing config data and other stuff thats not prepared yet
            if (this.isBenchmarking) {
                const html_data = "<!DOCTYPE html><html lang=\"en\"><head><meta charset=\"UTF-8\"><meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0\"><title>Document</title></head><body><div id=\"root\"></div></body></html>";
                this.cluster.queue({ slug, html_data });
                return;
            }

            // Here we need to calulate our HTML file... So its passed directly to the browser later
            /*
            [
                {
                uuid: 'D1BN-E0X5ZJxwFjETUa1g9mh4H9Y0yx_A7YIEN7lf8M8wcX3rXgrY37pGVgPT93oEZG_SVZkk_fR7w',
                server: 'euw'
                }
            ]
            */

            // Load the profiles from GQL
            const summoner_profiles_qgl = await fetchSummoner(puuid_array);

            // Itterate over the opject and check if any protected values is true
            Object.keys(summoner_profiles_qgl).forEach((key) => {
                if (summoner_profiles_qgl[key]?.protected) {
                    throw new Error("Protected Summoner");
                }
            });

            // Combine the puuid_array with the summoner_profiles_qgl
            /** @type {import('@lib/proto.js').CombinedSummonerProfile} */
            const summoner_request_gql = puuid_array.map((request, index) => {
                const profileKey = `profile${index}`;
                const profileData = summoner_profiles_qgl[profileKey];
                return { ...request, ...profileData };
            });

            const fallback_translations = this.config.LANGUAGES["en-US"];
            const translations = this.config.LANGUAGES[lang];

            const summoner_profiles = await asyncLoadProfiles(summoner_request_gql);
            let summoner_profiles_preMerge = summoner_profiles.map((sum) => {
                return sum.data;
            });

            // General Cards Data
            const card_raw_data = extractProtoData(mergeProtos(summoner_profiles_preMerge));
            card_raw_data.profileIcon_base64 = await getProfileWebPImageAsBase64(summoner_request_gql[0].profileIconId);

            // Alternative Cards Data
            if (design === "alternative") {
                const raw = mergeProtos(summoner_profiles_preMerge);
                card_raw_data.alternativ_all = extractProtoData(queueGroupMerger(raw, [3, 6, 8]));
                // calculate Alternative Cards gamemode Data
                card_raw_data.alternative_data_obj = {}
                card_raw_data.alternative_data_obj['urf'] = extractProtoData(queueGroupMerger(raw, [3]))
                card_raw_data.alternative_data_obj['arena'] = extractProtoData(queueGroupMerger(raw, [6]))
                card_raw_data.alternative_data_obj['nexus'] = extractProtoData(queueGroupMerger(raw, [8]))
            }

            // Add Username to the card_raw_data
            card_raw_data.username_main = `${summoner_request_gql[0].name}<span class="tagLine">#${summoner_request_gql[0].tagLine}</span>`;
            // Collect all usernames in a array
            const username_list = summoner_request_gql.map((summoner) => {
                return `${summoner.name}<span class="tagLine">#${summoner.tagLine}</span>`;
            });
            // Remove the first username from the array if we got more than 1
            if (username_list.length > 0) {
                username_list.shift();
            }

            const username_moreString = username_list.length === 1 ? `${username_list[0]}` : `${username_list[0]} and ${username_list.length - 1} more...`;
            card_raw_data.username_more = username_list.length > 0 ? username_moreString : "";

            // Add (Fallback) Language to the template data
            Object.keys(fallback_translations).forEach(key => {
                card_raw_data[`lang_${key}`] = fallback_translations[key];
            });
            // Add Language to the template data
            Object.keys(translations).forEach(key => {
                if (translations[key] !== "") card_raw_data[`lang_${key}`] = translations[key];
            });

            // Generate Localized Object for time functions
            const time_localized = {
                "days": card_raw_data[`lang_time_days`],
                "hours": card_raw_data[`lang_time_hours`],
                "minutes": card_raw_data[`lang_time_minutes`],
                "seconds": card_raw_data[`lang_time_seconds`],
            }

            /*
            const splitMe = process.me.split(".");
            card_raw_data.nodeID = crypto.createHash('sha256').update(String(`${splitMe[1]}${splitMe[2]}`)).digest('hex');
            */
            //console.log(card_raw_data.nodeID)

            /* ---------------------------------------------------------------------------------- */
            /* ------------------------------- Champion Processing ------------------------------ */

            // Artifically limit to 6 champions total
            // card_raw_data.champions = card_raw_data.champions.slice(0, 6);

            const champs_amount = card_raw_data.champions.length;

            // Add Skin and Champion ID to the card_raw_data (This is only done to the first 6 champions)
            const champions = card_raw_data.champions.slice(0, 6);
            // List all champ keys that should be beautified
            const beautifyChampValues = ['totalGames', 'totalWins', 'kills', 'deaths', 'assists', 'totalDamageTaken', 'cs', 'totalSeconds']
            const beautifyGamemoreValues = ['totalGames', 'totalWins', 'totalKills', 'totalDeaths', 'totalAssists', 'totaltotalDamageTaken', 'totalCS', 'totalSeconds']
            const obj_champions = {};

            if (design === "dark") {

                await Promise.all(champions.map(async (champ, index) => {
                    const champ_template_data = champ;

                    // Add (Fallback) Language to the template data
                    Object.keys(fallback_translations).forEach(key => {
                        champ_template_data[`lang_${key}`] = fallback_translations[key];
                    });
                    // Add Language to the template data
                    Object.keys(translations).forEach(key => {
                        champ_template_data[`lang_${key}`] = translations[key];
                    });

                    // Delete pentaKills if they are 0
                    if (champ_template_data.pentaKills === 0) {
                        champ_template_data.lang_champ_pentaKills = " ​ ";
                    }

                    // Singular pentaKills if they are 1
                    if (champ_template_data.pentaKills === 1) {
                        champ_template_data.lang_champ_pentaKills = champ_template_data[`lang_champ_pentaKill`];
                    }

                    const championData = getChampion(champ.champId);
                    const skin = getSkinById(champ.champId);
                    champ_template_data.img_tile = cupcake(`champion/tiles/${championData.id}_0`);
                    champ_template_data.img_tile_base64 = await getWebPImageAsBase64(champ_template_data.img_tile);
                    champ_template_data.img_splash = cupcake(`champion/splash/${championData.id}_${skin}`);
                    champ_template_data.img_splash_base64 = await getWebPImageAsBase64(champ_template_data.img_splash);
                    champ_template_data.playtime_humanformat = beautifySecondsToFixedString(champ.totalSeconds, true, time_localized);
                    champ_template_data.winrate = champ.totalGames > 0 ? (((champ.totalWins ?? 0) / (champ.totalGames ?? 1)) * 100).toFixed(1) : 0;
                    champ_template_data.kda = champ.deaths > 0 ? (champ.kills > 0 ? ((champ.kills + champ.assists) / champ.deaths).toFixed(1) : (champ.kills + champ.assists)) : 0;

                    // Check if it's the first champion (To get the centered image for header)
                    if (index === 0) {
                        card_raw_data.header_img_centered = cupcake(`champion/centered/${championData.id}_${skin}`);
                        card_raw_data.header_img_centered_base64 = await getWebPImageAsBase64(card_raw_data.header_img_centered);
                    }

                    beautifyChampValues.forEach((key) => {
                        champ_template_data[key] = shortenNumberByLocale(champ_template_data[key], translations);
                    });

                    obj_champions[`champ${index}`] = template(this.card_templates[design]['champ_full'], champ_template_data);
                }));

                if (typeof card_raw_data.header_img_centered === "undefined") {
                    card_raw_data.header_img_centered = cupcake(`champion/centered/Jinx_4`);
                    card_raw_data.header_img_centered_base64 = await getWebPImageAsBase64(card_raw_data.header_img_centered);
                }

                // Fill the champions object with more champions if we have less than 6 in obj_champions
                if (champs_amount < 6) {
                    for (let i = champs_amount; i < 6; i++) {
                        obj_champions[`champ${i}`] = template(this.card_templates[design]['champ_empty'], {});
                    }
                }

                // Merge the obj_champions together into ine html and then add it as obj_champions to the card_raw_data
                let html_champions = '';
                // Extract keys and sort them
                const sortedKeys = Object.keys(obj_champions).sort((a, b) => {
                    // Extract the numeric part from the champ keys and compare them
                    return parseInt(a.replace('champ', ''), 10) - parseInt(b.replace('champ', ''), 10);
                });

                for (const key of sortedKeys) {
                    html_champions += obj_champions[key];
                }

                card_raw_data.obj_champions = html_champions;

                // Flatten the object so we get every role directly in the root object
                Object.keys(card_raw_data.roles).forEach(role => {
                    // Add template keys for every role (games)
                    //card_raw_data.roles[role].games = 1
                    //card_raw_data.roles[role].wins = 0
                    card_raw_data[`lang_role_games_${role}`] = card_raw_data[`lang_role_games`].replace('{{role_games}}', `{{role_games_${role}}}`);
                    card_raw_data[`role_games_${role}`] = card_raw_data.roles[role].games;
                    // Add template keys for every role (winrate)
                    card_raw_data[`lang_role_winrate_${role}`] = card_raw_data.roles[role].games > 0 ? card_raw_data[`lang_role_winrate`].replace('{{role_winrate}}', `{{role_winrate_${role}}}`) : "   ";
                    card_raw_data[`role_winrate_${role}`] = card_raw_data.roles[role].games > 0 ? (card_raw_data.roles[role].wins / card_raw_data.roles[role].games * 100).toFixed(1) : "   "; // Now you know
                });
            }

            /* ------------------------------ Gamemodes Processing ------------------------------ */

            const obj_gamemodes = {};
            const gamemode_list = ['arena', 'urf', 'nexus'];
            const gamemode_bg = {
                "arena": "champion/centered/Yone_1",
                "urf": "champion/splash/Akali_61",
                "nexus": "champion/splash/Blitzcrank_29",
            }

            if (design === "alternative") {
                await Promise.all(gamemode_list.map(async (gamemode, index) => {
                    const gamemode_template_data = card_raw_data.alternative_data_obj[gamemode];

                    // Add (Fallback) Language to the template data
                    Object.keys(fallback_translations).forEach(key => {
                        gamemode_template_data[`lang_${key}`] = fallback_translations[key];
                    });
                    // Add Language to the template data
                    Object.keys(translations).forEach(key => {
                        gamemode_template_data[`lang_${key}`] = translations[key];
                    });

                    gamemode_template_data.modeName = gamemode_template_data[`lang_gamemode_${gamemode}`];
                    gamemode_template_data.modeTimestamp = gamemode_template_data[`lang_gamemode_${gamemode}_timestamp`];
                    gamemode_template_data.modeText = gamemode_template_data[`lang_gamemode_${gamemode}_text`];

                    if (gamemode_template_data.totalGames === 0) {
                        // Add Backround Image to the gamemode
                        gamemode_template_data.img_bg = cupcake(`profile/magic-bg`);
                        gamemode_template_data.img_bg_base64 = await getWebPImageAsBase64(gamemode_template_data.img_bg);
                        obj_gamemodes[index] = template(this.card_templates[design]['not_played'], gamemode_template_data);
                        index * 3; // This makes sure empty gamemodes are at the bottom
                        return;
                    }

                    // Add Backround Image to the gamemode
                    gamemode_template_data.img_bg = cupcake(gamemode_bg[gamemode]);
                    gamemode_template_data.img_bg_base64 = await getWebPImageAsBase64(gamemode_template_data.img_bg);

                    // Process the champions for the specific gamemode

                    // Add Skin and Champion ID to the card_raw_data (This is only done to the first 3 champions)
                    const champions = gamemode_template_data.champions.slice(0, 3);
                    const champ_obj = {};

                    await Promise.all(champions.map(async (champ, index) => {
                        const localClone = {};
                        Object.assign(localClone, gamemode_template_data);
                        // Delete pentaKills if they are 0
                        if (localClone[`champ_pentaKills`] === 0) {
                            localClone[`champ_lang_champ_pentaKills`] = " ​ ";
                        }

                        // Singular pentaKills if they are 1
                        if (champ.pentaKills === 1) {
                            localClone[`lang_champ_pentaKills`] = localClone[`champ${index}`][`lang_champ_pentaKill`];
                        }

                        const championData = getChampion(champ.champId);
                        const skin = getSkinById(champ.champId);
                        localClone[`champ_img_tile`] = cupcake(`champion/tiles/${championData.id}_0`);
                        localClone[`champ_img_tile_base64`] = await getWebPImageAsBase64(localClone[`champ_img_tile`]);
                        localClone[`champ_img_splash`] = cupcake(`champion/splash/${championData.id}_${skin}`);
                        localClone[`champ_img_splash_base64`] = await getWebPImageAsBase64(localClone[`champ_img_splash`]);
                        localClone[`champ_playtime_humanformat`] = beautifySecondsToFixedString(champ.totalSeconds, true, time_localized);
                        localClone[`champ_winrate`] = champ.totalGames > 0 ? (((champ.totalWins ?? 0) / (champ.totalGames ?? 1)) * 100).toFixed(1) : 0;
                        localClone[`champ_kda`] = champ.deaths > 0 ? (champ.kills > 0 ? ((champ.kills + champ.assists) / champ.deaths).toFixed(1) : (champ.kills + champ.assists)) : 0;

                        // Add template keys for the champions based on the gamemode
                        switch (gamemode) {
                            case "arena":
                                localClone[`champ_main_value`] = template(gamemode_template_data['lang_champ_kills'], champ);
                                break;
                            case "urf":
                                localClone[`champ_main_value`] = template(gamemode_template_data['lang_champ_kills'], champ);
                                break;
                            case "nexus":
                                localClone[`champ_main_value`] = template(gamemode_template_data['lang_champ_kills'], champ);
                                break;
                        }
                        

                        champ_obj[index] = template(this.card_templates[design]['champions'], localClone);
                    }));

                    // Sort the champ_obj by the index
                    const sortedChamps = Object.keys(champ_obj).sort((a, b) => {
                        // Extract the numeric part from the champ keys and compare them
                        return parseInt(a, 10) - parseInt(b, 10);
                    });
                    // Merge the champ_obj together into ine html and then add it as obj_champions to the card_raw_data
                    let html_champions = '';
                    for (const key of sortedChamps) {
                        html_champions += champ_obj[key];
                    }

                    gamemode_template_data.obj_champions = html_champions;

                    // Add template keys for the gamemode based on the gamemode
                    gamemode_template_data.playtime_humanformat = beautifySecondsToFixedString(gamemode_template_data.totalSeconds, true, time_localized);
                    gamemode_template_data.winrate = gamemode_template_data.totalGames > 0 ? (((gamemode_template_data.totalWins ?? 0) / (gamemode_template_data.totalGames ?? 1)) * 100).toFixed(1) : 0;
                    // Beautify the gamemore values
                    beautifyGamemoreValues.forEach((key) => {
                        gamemode_template_data[key] = shortenNumberByLocale(gamemode_template_data[key], translations);
                    });

                    obj_gamemodes[index] = template(this.card_templates[design][gamemode], gamemode_template_data);
                }));

                // Merge the obj_gamemodes together into ine html and then add it as obj_gamemodes to the card_raw_data
                let html_gamemodes = '';
                // Extract keys and sort them
                const sortedKeys = Object.keys(obj_gamemodes).sort((a, b) => {
                    // Extract the numeric part from the champ keys and compare them
                    return parseInt(a, 10) - parseInt(b, 10);
                });

                for (const key of sortedKeys) {
                    html_gamemodes += obj_gamemodes[key];
                }

                card_raw_data.obj_gamemodes = html_gamemodes;

                
                card_raw_data.header_img_centered = cupcake(`profile/magic-bg`);
                card_raw_data.header_img_centered_base64 = await getWebPImageAsBase64(card_raw_data.header_img_centered);
            }

            /* ---------------------------------------------------------------------------------- */
            /* ------------------------------ Custom Formated Stats ----------------------------- */

            if (design === "dark") {
                // Calculate Dark Global stats
                card_raw_data.totalWinrate = card_raw_data.totalGames > 0 ? (((card_raw_data.totalWins ?? 0) / (card_raw_data.totalGames ?? 1)) * 100).toFixed(2) : 0;
                card_raw_data.totalPlaytime_humanformat = beautifySecondsToFixedString(card_raw_data.totalSeconds ?? 0, false, time_localized);
                card_raw_data.totalSecondsSpentDead_humanformat = beautifySecondsToFixedString(card_raw_data.totalSecondsSpentDead ?? 0, true, time_localized);
                card_raw_data.totalDamage_humanformat = shortenNumberByLocale(card_raw_data.totalDamage ?? 0, translations);
                card_raw_data.totalKills_humanformat = shortenNumberByLocale(card_raw_data.totalKills ?? 0, translations);
                card_raw_data.totalCS_humanformat = shortenNumberByLocale(card_raw_data.totalCS ?? 0, translations);
                card_raw_data.totalGoldEarned_humanformat = shortenNumberByLocale(card_raw_data.totalGoldEarned ?? 0, translations);
                card_raw_data.totalButtonsPressed_humanformat = shortenNumberByLocale(card_raw_data.totalButtons ?? 0, translations);
                card_raw_data.totalChampionsPlayed = champs_amount;
            }

            if (design === "alternative") {
                card_raw_data.totalAlternativeGames = card_raw_data.alternativ_all.totalGames
                card_raw_data.totalAlternativeWinrate = card_raw_data.alternativ_all.totalGames > 0 ? (((card_raw_data.alternativ_all.totalWins ?? 0) / (card_raw_data.alternativ_all.totalGames ?? 1)) * 100).toFixed(2) : 0;
                card_raw_data.totalAlternativePlaytime_humanformat = beautifySecondsToFixedString(card_raw_data.alternativ_all.totalSeconds ?? 0, false, time_localized);
                card_raw_data.totalAlternativeSecondsSpentDead_humanformat = beautifySecondsToFixedString(card_raw_data.alternativ_all.totalSecondsSpentDead ?? 0, true, time_localized);
                card_raw_data.totalAlternativeDamage_humanformat = shortenNumberByLocale(card_raw_data.alternativ_all.totalDamage ?? 0, translations);
                card_raw_data.totalAlternativeKills_humanformat = shortenNumberByLocale(card_raw_data.alternativ_all.totalKills ?? 0, translations);
                card_raw_data.totalAlternativeCS_humanformat = shortenNumberByLocale(card_raw_data.alternativ_all.totalCS ?? 0, translations);
                card_raw_data.totalAlternativeGoldEarned_humanformat = shortenNumberByLocale(card_raw_data.alternativ_all.totalGoldEarned ?? 0, translations);
                card_raw_data.totalAlternativeButtonsPressed_humanformat = shortenNumberByLocale(card_raw_data.alternativ_all.totalButtons ?? 0, translations);
            }

            delete card_raw_data.roles;
            delete card_raw_data.champions;

            // Replace all {{Variables}} with the corresponding value of the Key
            const html_data = template(this.card_templates[design]['index'], card_raw_data);

            //console.log(card_raw_data)

            //const html_data = "<!DOCTYPE html><html lang=\"en\"><head><meta charset=\"UTF-8\"><meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0\"><title>Document</title></head><body><div id=\"root\"></div></body></html>";
            this.cluster.queue({ slug, html_data });

        } catch (error) {
            this.screenshot_failed++;
            this.screenshot_counter++;
            console.log(error)
            process.log.error(error);
            this.send("send:pic_err", {
                "slug": slug,
                "error": "General html Generation Error"
            });
        }
    }

    /**
     * Await for the cluster to be idle
     * @returns 
     */
    async idle() {
        await this.cluster.idle();
        return;
    }

    /**
     * Close the cluster
     * @param {boolean} live - When false its benchmarking
     * @returns
     */
    async end(live = false) {
        // Quit the app if it takes longer than it should
        if (live) {
            setTimeout(() => {
                process.log.error("Puppeteer cluster didn´t close in time. Exiting anyway.")
                process.exit(1);
            }, (Number(this.config.MAX_QUEUE_TIME) + 5) * 1000);
        }

        process.log.system("Waiting for Puppeteer Cluster to be idle");
        await this.cluster.idle();
        process.log.system("Closing Puppeteer Cluster");
        await this.cluster.close();

        clearInterval(this.some_Interval);

        this.cluster = null;
        return;
    }

    /**
     * Keeps the passed array at a certain length (100 times prope_interval)
     * @param {Array} array 
     */
    #shiftArrayAndCut = (array) => {
        const length = 60;
        if (array.length > length) {
            array.shift();
        }
    }

    findChromiumProcesses() {
        return new Promise((resolve, reject) => {
            let command;
            let parseOutput;

            if (os.platform() === 'win32') {
                // Windows
                command = 'tasklist /FI "IMAGENAME eq chrome.exe"';
                parseOutput = (stdout) => stdout.split('\n')
                    .filter(line => line.startsWith('chrome.exe'))
                    .map(line => {
                        const parts = line.trim().split(/\s+/);
                        return {
                            pid: parts[1],
                            memUsage: parts[4] / 1024,
                        };
                    });
            } else {
                // Linux/macOS
                command = 'ps aux | grep -i chromium';
                parseOutput = (stdout) => {
                    const totalRAM = os.totalmem() / 1024 / 1024; // Convert to MB
                    return stdout
                        .split('\n')
                        .filter(line => line.toLowerCase().includes('chromium'))
                        .map(line => {
                            const parts = line.trim().split(/\s+/);
                            const memPercentage = parseFloat(parts[3]);
                            const memUsageMB = (totalRAM * memPercentage) / 100; // Convert percentage to MB
                            return {
                                pid: parts[1],
                                memUsage: memUsageMB.toFixed(2), // Format to MB
                            };
                        });
                };
            }

            exec(command, (error, stdout, stderr) => {
                if (error) {
                    return reject(error);
                }
                resolve(parseOutput(stdout));
            });
        });
    }

    /**
     * Wipe base64 cache of images, they will be requested from cupcake again
     */
    wipecache() {
        clearimageCache();
    };

    get getMAX_CONCURRENCY() {
        return this.MAX_CONCURRENCY;
    }

    get getCluster() {
        return this.cluster;
    }

    get getchrome_memory_usage() {
        return this.chrome_memory_usage
    }

    get getscreenshot_rendertimes() {
        return this.screenshot_rendertimes;
    }

    get getscreenshot_renderd() {
        return this.screenshot_renderd;
    }

    get getscreenshot_failed() {
        return this.screenshot_failed;
    }

    get getrunsPerS_array() {
        return this.runsPerS_array;
    }

    get getimgCache_size() {
        return this.imgCache_size;
    }

    get getserved_mebibytes() {
        return this.served_mebibytes;
    }
}

module.exports = {
    PuppeteerClusterManager
};